//=========================================================
//  MusE
//  Linux Music Editor
//    $Id: synth.cpp,v 1.43 2004/07/14 15:27:26 wschweer Exp $
//  (C) Copyright 2000-2003 Werner Schweer (ws@seh.de)
//=========================================================

#include "config.h"
#include <sys/wait.h>
#include <signal.h>
#include <pthread.h>
#include <unistd.h>
#include <errno.h>
#include <sys/mman.h>
#include <vector>
#include <fcntl.h>

#include <qdir.h>
#include <dlfcn.h>
#include <qpopupmenu.h>

#ifdef VST_SUPPORT
#include <fst.h>
#include <vst/aeffectx.h>
#endif

#include "app.h"
#include "synth.h"
#include "xml.h"
#include "midi.h"
#include "midiport.h"
#include "mididev.h"
#include "libsynti/mess.h"
#include "song.h"
#include "audio.h"
#include "event.h"
#include "mpevent.h"
#include "audio.h"
#include "midiseq.h"
#include "midictrl.h"

std::vector<Synth*> synthis;  // array of available synthis

extern void connectNodes(AudioTrack*, AudioTrack*);
#ifdef VST_SUPPORT
extern long vstHostCallback (AEffect*, long, long, long, void*, float);
#endif

const char* Synth::description() const
      {
      if (descr)
            return descr->description;
      return "";
      }

const char* Synth::version() const
      {
      if (descr)
            return descr->version;
      return "";
      }

bool SynthI::guiVisible() const
      {
      if (_mess)
            return _mess->guiVisible();
      return _guiVisible;
      }

void SynthI::showGui(bool v)
      {
      if (v == guiVisible())
            return;
      if (_mess)
            _mess->showGui(v);
#ifdef VST_SUPPORT
      else {
            if (v)
                  fst_run_editor(_fst);
            else
                  fst_destroy_editor(_fst);
            _guiVisible = v;
            }
#endif
      }

bool SynthI::hasGui() const
      {
      if (_mess)
            return _mess->hasGui();
#ifdef VST_SUPPORT
      else
            return _fst->plugin->flags & effFlagsHasEditor;
#endif
      return false;
      }

MidiPlayEvent SynthI::receiveEvent()
      {
      if (_mess)
            return _mess->receiveEvent();
      return MidiPlayEvent();
      }

int SynthI::eventsPending() const
      {
      if (_mess)
            return _mess->eventsPending();
      return 0;
      }

void SynthI::getGeometry(int* x, int* y, int* w, int* h) const
      {
      if (_mess)
            _mess->getGeometry(x, y, w, h);
      }

void SynthI::setGeometry(int x, int y, int w, int h)
      {
      if (_mess)
            _mess->setGeometry(x, y, w, h);
      }

//---------------------------------------------------------
//   findSynth
//    search for synthesizer base class
//---------------------------------------------------------

static Synth* findSynth(const QString& sclass)
      {
      for (std::vector<Synth*>::iterator i = synthis.begin();
         i != synthis.end(); ++i) {
            if ((*i)->baseName() == sclass)
                  return *i;
            }
      printf("synthi class <%s> not found\n", sclass.latin1());
      return 0;
      }

//---------------------------------------------------------
//   createSynthI
//    create a synthesizer instance of class "label"
//---------------------------------------------------------

static SynthI* createSynthI(const QString& sclass)
      {
      Synth* s = findSynth(sclass);
      SynthI* si = 0;
      if (s) {
            si = new SynthI();
            QString n;
            n.setNum(s->instances());
            QString instance_name = s->baseName() + "-" + n;
            if (si->init(s, instance_name)) {
                  delete s;
                  return 0;
                  }
            }
      else
            printf("synthi class <%s> not found\n", sclass.latin1());
      return si;
      }

//---------------------------------------------------------
//   Synth
//---------------------------------------------------------

Synth::Synth(const QFileInfo& fi, bool type)
   : info(fi)
      {
      descr      = 0;
      fstHandle  = 0;
      vstType    = type;
      _instances = 0;
      }

//---------------------------------------------------------
//   incInstances
//---------------------------------------------------------

void Synth::incInstances(int val)
      {
      _instances += val;
      if (_instances == 0 && fstHandle) {
#ifdef VST_SUPPORT
            fst_unload(fstHandle);
#endif
            fstHandle = 0;
            }
      }

//---------------------------------------------------------
//   instantiate
//---------------------------------------------------------

void* Synth::instantiate()
      {
      ++_instances;
      QString n;
      n.setNum(_instances);
      QString instanceName = baseName() + "-" + n;
      doSetuid();
      const char* path = info.filePath().latin1();

      if (!vstType && descr == 0) {
            // load Synti dll
            void* handle = dlopen(path, RTLD_NOW);
            if (handle == 0) {
                  fprintf(stderr, "Synth::instantiate: dlopen(%s) failed: %s\n",
                     path, dlerror());
                  undoSetuid();
                  return 0;
                  }
            typedef const MESS* (*MESS_Function)();
            MESS_Function msynth = (MESS_Function)dlsym(handle, "mess_descriptor");

            if (!msynth) {
                  const char *txt = dlerror();
                  if (txt) {
                        fprintf(stderr,
                           "Unable to find msynth_descriptor() function in plugin "
                           "library file \"%s\": %s.\n"
                           "Are you sure this is a MESS plugin file?\n",
                           info.filePath().ascii(), txt);
                        undoSetuid();
                        return 0;
                        }
                  }
            descr = msynth();
            if (descr == 0) {
                  fprintf(stderr, "Synth::instantiate: no MESS descr found\n");
                  undoSetuid();
                  return 0;
                  }
            }
#ifdef VST_SUPPORT
      else if (vstType && fstHandle == 0) {
            fstHandle = fst_load(path);
            if (fstHandle == 0) {
                  printf("Synth::instantiate: cannot load vst plugin %s\n", path);
                  undoSetuid();
                  return 0;
                  }
            }
      if (vstType) {
            FST* fst = fst_instantiate(fstHandle, vstHostCallback, 0);
            if (fst == 0) {
                  printf("Synth::instantiate:: cannot instantiate plugin %s\n", path);
                  undoSetuid();
                  return 0;
                  }
            AEffect* plugin = fst->plugin;
            plugin->dispatcher (plugin, effMainsChanged, 0, 1, 0, 0.0f);

            /* set program to zero */

            plugin->dispatcher (plugin, effSetProgram, 0, 0, NULL, 0.0f);

            if (fst_run_editor(fst)) {
                  printf("Synth::instantiate: cannot create gui");
                  undoSetuid();
                  return 0;
                  }
            int vst_version = plugin->dispatcher (plugin, effGetVstVersion, 0, 0, NULL, 0.0f);

            if (vst_version >= 2) {
                  // should we send it VST events (i.e. MIDI)

//                  if ((plugin->flags & effFlagsIsSynth) ||
//                     (plugin->dispatcher(plugin, effCanDo, 0, 0,(void*) "receiveVstEvents", 0.0f) > 0)) {
//                        }
                  }
            // manage_vst_plugin(xxx);
            undoSetuid();
            return fst;
            }
      else
#endif
            {
            Mess* mess = descr->instantiate(sampleRate, muse, instanceName.latin1());
            undoSetuid();
            return mess;
            }
      }

//---------------------------------------------------------
//   SynthI
//---------------------------------------------------------

SynthI::SynthI()
   : AudioTrack(AUDIO_SOFTSYNTH)
      {
      synthesizer = 0;
      _mess       = 0;
      _fst        = 0;
      _rwFlags    = 1;
      _openFlags  = 1;
      _guiVisible = false;
      setVolume(1.0);
      setPan(0.0);
      }

void SynthI::setName(const QString& s)
      {
      AudioTrack::setName(s);
      MidiDevice::setName(s);
      }

//---------------------------------------------------------
//   init
//    returns false on success
//---------------------------------------------------------

bool SynthI::init(Synth* s, const QString& instanceName)
      {
      synthesizer = s;
#ifdef VST_SUPPORT
      if (s->isVst()) {
            _fst = (_FST*)synthesizer->instantiate();
            if (hasGui())
                  _guiVisible = true;
            }
      else
#endif
            _mess = (Mess*)synthesizer->instantiate();
      if (_mess == 0 && _fst == 0)
            return true;

      setName(instanceName);    // set midi device name
      setIName(instanceName);   // set instrument name

      if (!s->isVst()) {
            //---------------------------------------------------
            //  read available controller from synti
            //---------------------------------------------------

            int id = 0;
            MidiControllerList* cl = MidiInstrument::controller();
            for (;;) {
                  const char* name;
                  int ctrl;
                  int min;
                  int max;
                  id = _mess->getControllerInfo(id, &name, &ctrl, &min, &max);
                  if (id == 0)
                        break;
                  MidiController* c = new MidiController(QString(name), ctrl, min, max, 0);
                  cl->push_back(c);
                  }

            AudioTrack::setChannels(_mess->channels());

            EventList* iel = midiState();
            if (!iel->empty()) {
                  for (iEvent i = iel->begin(); i != iel->end(); ++i) {
                        Event ev = i->second;
                        MidiPlayEvent pev(0, 0, 0, ev);
                        if (putEvent(pev))
                              return false;   // try later
                        }
                  iel->clear();
                  }
            }
#ifdef VST_SUPPORT
      else {
            AEffect* plugin = _fst->plugin;
            setChannels(plugin->numOutputs);

            // initialize parameter:
            int idx = 0;
            for (std::vector<float>::iterator i = initParams.begin(); i != initParams.end(); ++i, ++idx)
                  plugin->setParameter(plugin, idx, *i);
            }
#endif
      return false;
      }

//---------------------------------------------------------
//   SynthI::deactivate
//---------------------------------------------------------

void SynthI::deactivate2()
      {
      removeMidiInstrument(this);
      midiDevices.remove(this);
      if (midiPort() != -1) {
            // synthi is attached
            midiPorts[midiPort()].setMidiDevice(0);
            }
      }

void SynthI::deactivate3()
      {
      if (_mess) {
            delete _mess;
            _mess = 0;
            }
#ifdef VST_SUPPORT
      if (_fst) {
            if (_guiVisible)
                  fst_destroy_editor(_fst);
            fst_close(_fst);
            _fst = 0;
            }
#endif
      synthesizer->incInstances(-1);
      }

//---------------------------------------------------------
//   ~SynthI
//---------------------------------------------------------

SynthI::~SynthI()
      {
      deactivate2();
      deactivate3();
      }

//---------------------------------------------------------
//   initMidiSynth
//    search for software synthis and advertise
//---------------------------------------------------------

void initMidiSynth()
      {
      QString s = museGlobalLib + "/synthi";

      QDir pluginDir(s, QString("*.so"), QDir::Files);
      if (debugMsg)
            printf("searching for software synthesizer in <%s>\n", s.latin1());
      if (pluginDir.exists()) {
            const QFileInfoList* list = pluginDir.entryInfoList();
            QFileInfoListIterator it(*list);
            QFileInfo* fi;
            while((fi = it.current())) {
                  synthis.push_back(new Synth(*fi, false));
                  ++it;
                  }
            if (debugMsg)
                  printf("%d soft synth found\n", synthis.size());
            }
      }

//---------------------------------------------------------
//   createSynthI
//    create a synthesizer instance of class "label"
//---------------------------------------------------------

SynthI* Song::createSynthI(const QString& sclass)
      {
      SynthI* si = ::createSynthI(sclass);
      insertTrack1(si, -1);
      msgInsertTrack(si, -1, true);       // add to instance list
      insertTrack3(si, -1);

      OutputList* ol = song->outputs();
      // add default route to master (first audio output)
      if (!ol->empty()) {
            AudioOutput* ao = ol->front();
            audio->msgAddRoute(Route(si, -1), Route(ao, -1));
            }
      return si;
      }

//---------------------------------------------------------
//   write
//---------------------------------------------------------

void SynthI::write(int level, Xml& xml) const
      {
      xml.tag(level++, "SynthI");
      AudioTrack::writeProperties(level, xml);
      xml.strTag(level, "class", synth()->baseName());

      //---------------------------------------------
      // if soft synth is attached to a midi port,
      // write out port number
      //---------------------------------------------

      if (midiPort() != -1)
            xml.intTag(level, "port", midiPort());

      if (hasGui()) {
            xml.intTag(level, "guiVisible", guiVisible());
            int x, y, w, h;
            w = 0;
            h = 0;
            getGeometry(&x, &y, &w, &h);
            if (h || w)
                  xml.qrectTag(level, "geometry", QRect(x, y, w, h));
            }

      //---------------------------------------------
      // dump current state of synth
      //---------------------------------------------

      int len = 0;
      const unsigned char* p;
      if (_mess) {
            _mess->getInitData(&len, &p);
            if (len) {
                  xml.tag(level++, "midistate");
                  xml.nput(level++, "<event type=\"%d\"", Sysex);
                  xml.nput(" datalen=\"%d\">\n", len);
                  xml.nput(level, "");
                  for (int i = 0; i < len; ++i) {
                        if (i && ((i % 16) == 0)) {
                              xml.nput("\n");
                              xml.nput(level, "");
                              }
                        xml.nput("%02x ", p[i] & 0xff);
                        }
                  xml.nput("\n");
                  xml.tag(level--, "/event");
                  xml.etag(level--, "midistate");
                  }
            }
#ifdef VST_SUPPORT
      else if (_fst) {
            AEffect* plugin = _fst->plugin;
            int params = plugin->numParams;
            for (int i = 0; i < params; ++i) {
                  float f = plugin->getParameter(plugin, i);
                  xml.floatTag(level, "param", f);
                  }
            }
#endif
      xml.etag(level, "SynthI");
      }

//---------------------------------------------------------
//   SynthI::read
//---------------------------------------------------------

void SynthI::read(Xml& xml)
      {
      QString sclass;
      int port = -1;
      bool startGui = false;
      QRect r;

      for (;;) {
            Xml::Token token = xml.parse();
            const QString tag = xml.s1();
            switch (token) {
                  case Xml::Error:
                  case Xml::End:
                        return;
                  case Xml::TagStart:
                        if (tag == "class")
                              sclass = xml.parse1();
                        else if (tag == "port")
                              port  = xml.parseInt();
                        else if (tag == "guiVisible")
                              startGui = xml.parseInt();
                        else if (tag == "midistate")
                              readMidiState(xml);
                        else if (tag == "param") {
                              float val = xml.parseFloat();
                              initParams.push_back(val);
                              }
                        else if (tag == "geometry")
                              r = readGeometry(xml, tag);
                        else if (AudioTrack::readProperties(xml, tag))
                              xml.unknown("softSynth");
                        break;
                  case Xml::TagEnd:
                        if (tag == "SynthI") {
                              Synth* s = findSynth(sclass);
                              if (s == 0)
                                    return;
                              if (init(s, name())) {
                                    return;
                                    }
                              song->insertTrack0(this, -1);
                              if (port != -1 && port < MIDI_PORTS)
                                    midiPorts[port].setMidiDevice(this);
                              showGui(startGui);
                              setGeometry(r.x(), r.y(), r.width(), r.height());
                              return;
                              }
                  default:
                        break;
                  }
            }
      }

//---------------------------------------------------------
//   getPatchName
//---------------------------------------------------------

const char* SynthI::getPatchName(int channel, int prog, MType type)
      {
      if (_mess)
            return _mess->getPatchName(channel, prog, type);
      return "";
      }

//---------------------------------------------------------
//   populatePatchPopup
//---------------------------------------------------------

void SynthI::populatePatchPopup(QPopupMenu* menu, int ch, MType)
      {
      menu->clear();
      const MidiPatch* mp = _mess->getPatchInfo(ch, 0);
      while (mp) {
            int id = ((mp->hbank & 0xff) << 16)
                      + ((mp->lbank & 0xff) << 8) + mp->prog;
            menu->insertItem(QString(mp->name), id);
            mp = _mess->getPatchInfo(ch, mp);
            }
      }

//---------------------------------------------------------
//   getData
//    - run MESS ports
//---------------------------------------------------------

bool SynthI::getData(unsigned pos, int ports, unsigned n, float** buffer)
      {
      int p = midiPort();
      MidiPort* mp = (p != -1) ? &midiPorts[p] : 0;
      MPEventList* el = playEvents();

      for (int i = 0; i < ports; ++i)
            memset(buffer[i], 0, n * sizeof(float));

      if (_mess) {
            iMPEvent i      = nextPlayEvent();
            int curPos      = pos;
            int endPos      = pos + n;
            int off         = pos;
            int frameOffset = audio->getFrameOffset();

            for (; i != el->end(); ++i) {
                  int frame = i->time() - frameOffset;
                  if (frame >= endPos)
                        break;
                  if (frame > curPos) {
                        if (frame < pos)
                              printf("should not happen: missed event %d\n", pos -frame);
                        else
                              _mess->process(buffer, curPos-pos, frame - curPos);
                        curPos = frame;
                        }
                  if (mp)
                        mp->sendEvent(*i);
                  else {
                        if (putEvent(*i))
                              break;
                        }
                  }
            if (endPos - curPos) {
                  _mess->process(buffer, curPos - off, endPos - curPos);
                  }
            setNextPlayEvent(i);
            }
#ifdef VST_SUPPORT
      else if (_fst) {
            AEffect* plugin = _fst->plugin;
            for (iMPEvent i = nextPlayEvent(); i != el->end(); ++i) {
                  if (mp)
                        mp->sendEvent(*i);
                  else {
                        if (putEvent(*i))
                              break;
                        }
                  }
            setNextPlayEvent(el->end());
            if (plugin->flags & effFlagsCanReplacing) {
                  plugin->processReplacing(plugin, 0, buffer, n);
                  }
            else {
                  plugin->process(plugin, 0, buffer, n);
                  }
            }
#endif
      return true;
      }

//---------------------------------------------------------
//   putEvent
//    return true on error (busy)
//---------------------------------------------------------

bool SynthI::putEvent(const MidiPlayEvent& ev)
      {
      if (midiOutputTrace)
            ev.dump();
      if (_mess)
            return _mess->processEvent(ev);
#ifdef VST_SUPPORT
      else {
            AEffect* plugin = _fst->plugin;
            static struct VstEvents events;
            static struct VstMidiEvent event;
            events.numEvents = 1;
            events.reserved  = 0;
            events.events[0] = (VstEvent*)(&event);

            event.type         = kVstMidiType;
            event.byteSize     = 24;
            event.deltaFrames  = 0;
            event.flags        = 0;
            event.detune       = 0;
            event.noteLength   = 0;
            event.noteOffset   = 0;
            event.reserved1    = 0;
            event.reserved2    = 0;
            event.noteOffVelocity = 0;
            switch (ev.type()) {
                  case ME_PITCHBEND:
                        {
                        int a = ev.dataA() + 8192;
                        int b = a >> 7;
                        event.midiData[0]  = (ev.type() | ev.channel()) & 0xff;
                        event.midiData[1]  = a & 0x7f;
                        event.midiData[2]  = b & 0x7f;
                        event.midiData[3]  = 0;
                        }
                        break;

                  case ME_CONTROLLER:
                  case ME_NOTEON:
                  default:
                        event.midiData[0]  = (ev.type() | ev.channel()) & 0xff;
                        event.midiData[1]  = ev.dataA() & 0xff;
                        event.midiData[2]  = ev.dataB() & 0xff;
                        event.midiData[3]  = 0;
                        break;
                  }
            int rv = plugin->dispatcher(plugin, effProcessEvents, 0, 0, &events, 0.0f);
            return false;
            }
#else
      return true;
#endif
      }

